// oauthorize.go
// sample to demonstrate using golang oauth2 and scs libraries
// to provide authorization for Single-Page webapps
package main

import (
	"bytes"
	"encoding/json"
	"fmt"
	"os"
	"strconv"
	"text/template"
	"time"

	"net/http"
	"net/url"

	"github.com/dchest/uniuri"
	"github.com/golang-jwt/jwt/v5"
	"github.com/gorilla/mux"

	"golang.org/x/oauth2"

	eventsource "tps-git.topcon.com/event-sourcing/event-source-go"
	eventstore "tps-git.topcon.com/event-sourcing/event-store-go"

	"tps-git.topcon.com/goutil/fatal-go"
	"tps-git.topcon.com/goutil/logger-go"
)

const (
	pingHost  = "qa-token.auth.topcon.com"
	trapiHost = "api-qa.topcon.com"
)

// OAuthorize: sample to demonstrate using golang oauth2 and scs libraries
// to provide authorization for Single-Page webapps
type OAuthorize struct {
	RedirectPort int
	ServerURL    string

	indexTemplate   *template.Template
	logoutTemplate  *template.Template
	oauth2Config    *oauth2.Config
	store           eventsource.Store
	verifierStore   eventsource.Store
	endSessionStore eventsource.Store
}

func NewOAuthorize(scheme, host string, redirectPort int) (oa *OAuthorize) {
	var (
		clientID     = os.Getenv("CLIENT_ID")
		clientSecret = os.Getenv("CLIENT_SECRET")
	)

	storeFactory, err := eventstore.StoreFactoryFromEnv("SESSION_FACTORY", "MEM")
	fatal.OnError(err)

	// convenience methods for making things:
	mustTemplate := func(key string) *template.Template {
		t, err := template.New(key).Parse(string(MustAsset(key)))
		fatal.OnError(err)
		return t
	}

	mustStore := func(key string) eventsource.Store {
		store, err := storeFactory.NewStore(key)
		fatal.OnError(err)
		return store
	}

	serverURL := fmt.Sprintf("%s://%s:%d", scheme, host, redirectPort)
	return &OAuthorize{
		RedirectPort:   redirectPort,
		ServerURL:      serverURL,
		indexTemplate:  mustTemplate("index.html"),
		logoutTemplate: mustTemplate("logout.html"),
		oauth2Config: &oauth2.Config{
			RedirectURL:  fmt.Sprintf("%s/auth.callback", serverURL),
			ClientID:     clientID,
			ClientSecret: clientSecret,
			Scopes:       []string{"openid", "topcon.ciam.phase.1.5"},
			Endpoint: oauth2.Endpoint{
				AuthURL:   fmt.Sprintf("https://%s/as/%s.oauth2", pingHost, "authorization"),
				TokenURL:  fmt.Sprintf("https://%s/as/%s.oauth2", pingHost, "token"),
				AuthStyle: oauth2.AuthStyleInHeader,
			},
		},
		store:           mustStore("oauthorize-session-store"),
		verifierStore:   mustStore("oauthorize-verifier-store"),
		endSessionStore: mustStore("oauthorize-endsession-store"),
	}
}

// private methods -------------------------------------------------------------

func (oa *OAuthorize) AddHandlers(router *mux.Router) {
	router.HandleFunc("/", oa.handleLogin)
	router.HandleFunc("/auth.callback", oa.handleCallback) // oauth callback
	router.HandleFunc("/auth.switch", oa.handleSwitch)     // change teams
	router.HandleFunc("/logout", oa.handleLogout)
	router.HandleFunc("/auth.token", oa.handleAccessToken)
}

// handlers and an internal method to end a session

func (oa *OAuthorize) endSession(r *http.Request) {
	sessionId := r.Header.Get("session")
	if sessionId == "" {
		if s := r.FormValue("__session"); s != "" {
			if b, _ := oa.endSessionStore.Get(s); b != nil {
				sessionId = string(b)
			}
		}
	}
	if sessionId != "" {
		logger.Notice("removing session %s", sessionId)
		oa.store.Remove(sessionId)
	}
}

func (oa *OAuthorize) handleLogin(w http.ResponseWriter, r *http.Request) {
	endpoint := oa.makeAuthCodeURL(false)
	logger.Tracef("redirect to %s", endpoint)
	http.Redirect(w, r, endpoint, http.StatusTemporaryRedirect)
}

func (oa *OAuthorize) handleSwitch(w http.ResponseWriter, r *http.Request) {
	oa.endSession(r)
	endpoint := oa.makeAuthCodeURL(true)
	logger.Tracef("redirect to %s", endpoint)
	http.Redirect(w, r, endpoint, http.StatusTemporaryRedirect)
}

func (oa *OAuthorize) handleLogout(w http.ResponseWriter, r *http.Request) {
	var err error
	defer func() {
		if err != nil {
			http.Error(w, err.Error(), http.StatusInternalServerError)
		}
	}()

	oa.endSession(r)
	var stage int
	sstage := r.FormValue("__stage")
	if sstage != "" {
		stage, err = strconv.Atoi(sstage)
		if err != nil {
			return
		}
	}

	nextStageURL := func() string { return fmt.Sprintf(`%s/logout?__stage=%d`, oa.ServerURL, stage+1) }
	switch stage {
	default:
		err = fmt.Errorf("unexpected stage value %d", stage)
		return
	case 0:
		// this stage has removed the session value, provided it got passed correctly
		http.Redirect(w, r, nextStageURL(), http.StatusSeeOther)
	case 1:
		// call PingFederate's startSLO path.
		// this removes team/profile settings
		queryParams := url.Values{}
		queryParams.Add("TargetResource", nextStageURL())
		apiEndpoint := url.URL{Scheme: "https", Host: pingHost, Path: "/idp/startSLO.ping", RawQuery: queryParams.Encode()}
		http.Redirect(w, r, apiEndpoint.String(), http.StatusSeeOther)
	case 2:
		// call trapi's logout path
		// this removes the authorization at PingOne.
		queryParams := url.Values{}
		queryParams.Add("ping", "true")
		queryParams.Add("redirect", nextStageURL())
		apiEndpoint := url.URL{Scheme: "https", Host: trapiHost, Path: "/logout", RawQuery: queryParams.Encode()}
		http.Redirect(w, r, apiEndpoint.String(), http.StatusSeeOther)
	case 3:
		// final stage: render a "you are now logged out" page:
		oa.logoutTemplate.Execute(w, map[string]string{
			"serverURL": oa.ServerURL,
		})
		return
	}

}

func (oa *OAuthorize) makeAuthCodeURL(switching bool) string {
	verifier := oauth2.GenerateVerifier()
	verifierKey := uniuri.NewLen(40)
	oa.verifierStore.Set(verifierKey, []byte(verifier))
	endpoint := oa.oauth2Config.AuthCodeURL(verifierKey, oauth2.S256ChallengeOption(verifier), oauth2.AccessTypeOffline)
	if switching {
		endpoint += "&switch=bp"
	}
	return endpoint
}

func (oa *OAuthorize) handleCallback(w http.ResponseWriter, r *http.Request) {
	var err error
	defer func() {
		if err != nil {
			logger.Warnf("failed token exchange %s", err.Error())
			http.Error(w, err.Error(), 500)
		}
	}()

	state := r.FormValue("state")
	verifierBytes, _ := oa.verifierStore.Get(state)
	oa.verifierStore.Remove(state)
	verifier := string(verifierBytes)
	code := r.FormValue("code")
	token, err := oa.oauth2Config.Exchange(oauth2.NoContext, code, oauth2.VerifierOption(verifier))
	if err != nil {
		return
	}

	sessionId := uniuri.NewLen(48)
	err = oa.store.Set(sessionId, fatal.UnlessMarshal(token))
	if err != nil {
		return
	}

	endSessionId := uniuri.NewLen(48)
	oa.endSessionStore.Set(endSessionId, []byte(sessionId))

	oa.indexTemplate.Execute(w, map[string]string{
		"session":   sessionId,
		"switchURL": "/auth.switch?__session=" + endSessionId,
		"logoutURL": "/logout?__session=" + endSessionId,
	})
}

func (oa *OAuthorize) handleAccessToken(w http.ResponseWriter, r *http.Request) {
	var err error
	defer func() {
		if err != nil {
			http.Error(w, err.Error(), http.StatusUnauthorized)
		}
	}()

	sessionId := r.Header.Get("session")
	if err != nil {
		return
	}

	sessionBytes, err := oa.store.Get(sessionId)
	if err != nil {
		return
	}

	var ttoken oauth2.Token
	token := &ttoken
	decoder := json.NewDecoder(bytes.NewReader(sessionBytes))
	decoder.UseNumber()
	err = decoder.Decode(token)
	if err != nil {
		return
	}

	jwtExpiry := func(aJWT string) (exp jwt.NumericDate, err error) {
		var claims jwt.MapClaims = make(jwt.MapClaims)
		jwt.NewParser().ParseUnverified(aJWT, claims)

		ex, err := claims.GetExpirationTime()
		if err != nil {
			return
		}
		if ex == nil {
			err = fmt.Errorf("no expiry provided")
			return
		}
		return *ex, nil
	}

	exp, err := jwtExpiry(token.AccessToken)
	if err != nil {
		return
	}
	if !exp.After(time.Now()) {
		tokenSource := oa.oauth2Config.TokenSource(oauth2.NoContext, token)
		token, err = tokenSource.Token()
		if err != nil {
			return
		}
		err = oa.store.Set(sessionId, fatal.UnlessMarshal(token))
		if err != nil {
			return
		}
		exp, err = jwtExpiry(token.AccessToken)
		if err != nil {
			return
		}
	}

	jsonResp, _ := json.Marshal(map[string]string{
		"accessToken": token.AccessToken,
		"expiry":      strconv.FormatInt(exp.Unix(), 10),
	})
	w.Write(jsonResp)
	return
}
